function [snapshot_matrix, computation_time, u_ode45, t_ode45, computation_time_ode45, u_ode23, t_ode23, computation_time_ode23] = compute_truth_solution(fom, discretization, D, DD, const_stepsize)

% Computes the full order solution based on a self written time integration (implicit trapezoidal rule) of the equation fom.equation based on the parameters in the
% structures fom and discretization. Other inputs are finite difference matrices D and DD for the first and second
% spatial derivative, as well as a flag const_stepsize for the ode solvers.
% If more than three outputs are requested, this function also uses matlab ode45 and ode23 solvers to compute the full order solution.

equation = fom.equation;

dx = discretization.space.dx;
n_steps_space = discretization.space.n_steps;
L = discretization.space.length;

dt = discretization.time.dt;
n_steps_time = discretization.time.n_steps;

initial_condition = fom.initial_condition;
n_vars = fom.n_vars;

% space
space = dx * (1:n_steps_space);

% time
time = dt * (0:n_steps_time-1);

switch equation
    case 'advection_reaction_diffusion'
        u_0 = initial_condition(space, L);
        c = fom.parameter.c;
        mu = fom.parameter.mu;
        r = fom.parameter.r;
    case 'burgers'
        u_0 = initial_condition(space, L);
        mu = fom.parameter.mu;
    case 'wave_equation'
        u_0 = [initial_condition{1}(space, L), initial_condition{2}(space, L)];
        c = fom.parameter.c;
        rho = fom.parameter.rho;
end

% initialize snapshot matrix
snapshot_matrix = zeros(n_steps_space * n_vars, n_steps_time);
snapshot_matrix(:,1) = u_0';

% initialize solution
u = u_0';

% If the output is requested, compute also the ode45 and ode23 solution.
% Note that u_ode45 / u_ode23 is already a matrix, time by space, in contrast to the snapshot_matrix, with is space by time.

if (nargout > 2)
    u_ode45 = u;
    u_ode23 = u;
    
    opts = odeset('Refine', 1);
    switch equation
        case 'burgers'
        
            if const_stepsize
                tsize = discretization.time.dt * (0:1:discretization.time.n_steps-1);
            else
                tsize = [0,1];
            end
        
            tic;
            [t_ode45, u_ode45] = ode45(@(t, u) evaluate_burgers_right_hand_side(t, dx, u, D, DD, mu), tsize, u_ode45, opts);
            computation_time_ode45 = toc;
            
            tic;
            [t_ode23, u_ode23] = ode23(@(t, u)evaluate_burgers_right_hand_side(t, dx, u, D, DD, mu), tsize, u_ode23);
            computation_time_ode23 = toc;
            
            u_ode23 = u_ode23';
            u_ode45 = u_ode45';
            
        case 'advection_reaction_diffusion'
        
            if const_stepsize
                tsize = discretization.time.dt * (0:1:discretization.time.n_steps-1);
            else
                tsize = [0,1];
            end
        
            tic;
            [t_ode45, u_ode45] = ode45(@(t, u) evaluate_advection_diffusion_right_hand_side(t, dx, u, D, DD, c, mu, r), tsize, u_ode45, opts);
            computation_time_ode45 = toc;
            
            tic;
            [t_ode23, u_ode23] = ode23(@(t, u)evaluate_advection_diffusion_right_hand_side(t, dx, u, D, DD, c, mu, r), tsize, u_ode23);
            computation_time_ode23 = toc;
            
            u_ode23 = u_ode23';
            u_ode45 = u_ode45';
            
        case 'wave_equation'
        
            if const_stepsize
                tsize = discretization.time.dt * (0:1:discretization.time.n_steps-1);
            else
                tsize = [0,1];
            end
        
            tic;
            [t_ode45, u_ode45] = ode45(@(t, u) evaluate_wave_equation_right_hand_side(t, dx, u, D, c, rho), tsize, u_ode45, opts);
            computation_time_ode45 = toc;
            
            tic;
            [t_ode23, u_ode23] = ode23(@(t, u)evaluate_wave_equation_right_hand_side(t, dx, u, D, c, rho), tsize, u_ode23);
            computation_time_ode23 = toc;
            
            % Transpose ode solver solutions due to them being time by space.
            u_ode23 = u_ode23';
            u_ode45 = u_ode45';
    end
end

% Self written time-integration.

switch equation

    case 'advection_reaction_diffusion'
        tic;
        for i = 1:n_steps_time-1
            u = implicit_trapezoid(time(i), dt, @(t,u) evaluate_advection_diffusion_right_hand_side(t, dx, u, D, DD, c, mu, r), u);
            snapshot_matrix(:,i+1) = u;
        end
        computation_time = toc;
        
    case 'wave_equation'
        tic;
        for i = 1:n_steps_time-1
            u = implicit_trapezoid(time(i), dt, @(t,u)evaluate_wave_equation_right_hand_side(t, dx, u, D, c, rho), u);
            snapshot_matrix(:,i+1) = u;
        end
        computation_time = toc;
        
    case 'burgers'
        tic;
        for i = 1:n_steps_time-1
            u = implicit_trapezoid(time(i), dt, @(t,u) evaluate_burgers_right_hand_side(t, dx, u, D, DD, mu), u);
            snapshot_matrix(:,i+1) = u;
        end
        computation_time = toc;
        
    otherwise
        error(['The string value ', equation, ' does not specify a known example.']);
        
end

end
