function rom = create_rom_struct(equation, varargin)
 
    % Helper function that creates the struct array rom which stores all relevant parameters
    % for the simulation of the shifted POD rom.
 
    behaviour = 'multiplicative';

    % Check if varargin contains admissible name-value pairs.
    switch equation
        case 'advection_reaction_diffusion'
            admissible_names = {'parameter.c', 'parameter.mu', 'parameter.r', 'initial_condition'};
        case 'wave_equation'
            admissible_names = {'initial_condition', 'parameter.c', 'parameter.rho'};
        case 'burgers'
            admissible_names = {'initial_condition', 'parameter.mu'};
        otherwise
            error("%s is not a known example!", equation);
    end
    
    % If length(varargin) is not even, the user forgot a name or a value.
    if mod(length(varargin), 2) == 1
        error("Not enough input arguments: The number of variable input arguments is not even, indicating that there is either a name or a value missing in a name-value pair.");
    end
    
    % Check for non-admissible names and wrong values
    
    % Get the names from varargin
    names = varargin(1:2:end);
    % Get the corresponding values from varargin
    values = varargin(2:2:end);
    
    % Get the number of elements in the values 
    n_elements = zeros(1, length(values));
    for i = 1:length(values)
        n_elements(i) = length(values{i});
    end
    
    for i = 1:length(names)
    
        current_name = names{i};
        current_value = values{i};
        
        % If the name is not specified as string or char array, throw an error
        if ~ischar(current_name) && ~isstring(current_name)
            error("Names need to be specified as strings or character arrays!");
        end
        
        % Check if name is admissible
        is_admissible = false;
        for j = 1:length(admissible_names)
            if strcmp(current_name, admissible_names{j})
                is_admissible = true;
                break;
            end
        end
        
        if ~is_admissible
            error("%s is not an admissible name for a name-value pair!", current_name);
        end
        
        % Check for wrong values
        switch current_name
        
            case 'initial_condition'
            
                % Check if value is a cell of function handles or a single function handle
                if ~isa(current_value, 'function_handle') && ~iscell(current_value)
                    error("The value for %s must be a function handle or a cell of function handles or a cell of cells with function handles in case of the wave_equation!", current_name);
                end
                
                % If a cell is provided, check if all entries are function handles or cells of function handles (wave_equation)
                if iscell(current_value)
                    for i = 1:length(current_value)
                        if ~isa(current_value{i}, 'function_handle') && ~iscell(current_value{i})
                            error("The value for %s must be a function handle, a cell of function handles or a cell that contains cells of function handles in case of the wave_equation!", current_name);
                        end
                        if iscell(current_value{i})
                            for j = 1:length(current_value{i})
                                if ~isa(current_value{i}{j}, 'function_handle')
                                    error("The value for %s must be a function handle, a cell of function handles or a cell that contains cells of function handles in case of the wave_equation!");
                                end
                            end
                        end
                    end
                end
                
            case 'parameter.c'
                if ~isnumeric(current_value)
                    error("The value for %s must be a numeric array!", current_name);
                end
                
            case 'parameter.mu'
                if ~isnumeric(current_value)
                    error("The value for %s must be a numeric array!", current_name);
                end
                
            case 'parameter.r'
                if ~isnumeric(current_value)
                    error("The value for %s must be a numeric array!", current_name);
                end
                
            case 'parameter.rho'
                if ~isnumeric(current_value)
                    error("The value for %s must be a numeric array!", current_name);
                end
                
            otherwise
                error("%s is not an admissible name for a name-value pair!", current_name);
        end
        
    end
    
    %% Process the name-value pairs into a struct array, depending on behaviour
   
    % Compute the number of structs in the rom struct array
    if strcmp(behaviour, 'multiplicative')
        n_structs = 1;
        for i = 1:length(values)
            current_value = values{i};
            n_structs = n_structs * length(current_value);
        end
    end
    
    % Initialization of struct array
    for i = 1:n_structs
        rom(i) = struct();
    end
    
    switch length(n_elements)
        case 1
            indices = 1:n_structs;
        case 2
            [ind1, ind2] = ind2sub(n_elements, (1:n_structs));
            indices = [ind1; ind2];
        case 3
            [ind1, ind2, ind3] = ind2sub(n_elements, (1:n_structs));
            indices = [ind1; ind2; ind3];
        case 4
            [ind1, ind2, ind3, ind4] = ind2sub(n_elements, (1:n_structs));
            indices = [ind1; ind2; ind3; ind4];
        case 5
            [ind1, ind2, ind3, ind4, ind5] = ind2sub(n_elements, (1:n_structs));
            indices = [ind1; ind2; ind3; ind4; ind5];
        case 6
            [ind1, ind2, ind3, ind4, ind5, ind6] = ind2sub(n_elements, (1:n_structs));
    end
    
    % Write values to the structs
    for i = 1:length(names)
    
        field_name = names{i};
        field_values = values{i};
        
        for j = 1:n_structs
            switch field_name
                case 'initial_condition'
                    switch equation
                        case 'advection_reaction_diffusion'
                            rom(j).initial_condition = field_values{indices(i,j)};
                        case 'wave_equation'
                            for k = 1:length(field_values{indices(i,j)})
                                rom(j).initial_condition{k} = field_values{indices(i,j)}{k};
                            end
                        case 'burgers'
                            rom(j).initial_condition = field_values{indices(i,j)};
                    end
                case 'parameter.c'
                    rom(j).parameter.c = field_values(indices(i,j));
                case 'parameter.mu'
                    rom(j).parameter.mu = field_values(indices(i,j));
                case 'parameter.r'
                    rom(j).parameter.r = field_values(indices(i,j));
                case 'parameter.rho'
                    rom(j).parameter.rho = field_values(indices(i,j));
            end
        end
        
    end
    
    % Set default values if there are no optional parameters
    default_names = setdiff(admissible_names, names);
    for i = 1:length(default_names)
    
        field_name = default_names{i};
        
            for j = 1:n_structs
                switch field_name
                    case 'initial_condition'
                        switch equation
                            case 'advection_reaction_diffusion'
                                rom(j).initial_condition = @gaussian_pulse;
                            case 'wave_equation'
                                rom(j).initial_condition{1} = @gaussian_pulse;
                                rom(j).initial_condition{2} = @(x,L) zeros(1, length(x));
                            case 'burgers'
                                rom(j).initial_condition = @gaussian_pulse;
                        end
                    case 'parameter.c'
                        rom(j).parameter.c = 1;
                    case 'parameter.mu'
                        rom(j).parameter.mu = 0.002;
                    case 'parameter.r'
                        rom(j).parameter.r = 0;
                    case 'parameter.rho'
                        rom(j).parameter.rho = 1;
                end
                rom(j).equation = equation;
            end
            
    end

end
