function [POD_solution, reduced_state_matrix, computation_time, POD_solution_ode45, t_ode45, reduced_state_ode45, computation_time_ode45, POD_solution_ode23, t_ode23, reduced_state_ode23, computation_time_ode23] = POD_online(equation, modes, discretization, parameter, initial_condition, D, DD, const_stepsize)
    
    % Computes the POD ROM solution based on a self written time integration (implicit trapezoidal rule) of the equation 'equation' based on the parameters in the
    % structures discretization and parameter. Other inputs are finite difference matrices D and DD for the first and second
    % spatial derivative, as well as a flag const_stepsize for the ode solvers and of course the modes.
    % If more than three outputs are requested, this function also uses matlab ode45 and ode23 solvers to compute the full order solution.

    
    switch equation
        case 'advection_reaction_diffusion'
            n_vars = 1;
        case 'wave_equation'
            n_vars = 2;
        case 'burgers'
            n_vars = 1;
        otherwise
            error("%s is not a known example!", equation);
    end
    
%      c = parameter.c;
%      mu = parameter.mu;
%      r = parameter.r;
%      rho = parameter.rho;
    
%      D1 = parameter.discretization.D1;
%      D2 = parameter.discretization.D2;
    
    dx = discretization.space.dx;
    n_steps_space = discretization.space.n_steps;
    space = dx * (1:n_steps_space);
    space_length = discretization.space.length;
    
    dt = discretization.time.dt;
    n_steps_time = discretization.time.n_steps;
    time = dt * (0:1:n_steps_time-1);
    
    switch equation
    
        case 'advection_reaction_diffusion'
        
            c = parameter.c;
            mu = parameter.mu;
            r = parameter.r;
            
            right_hand_side_matrix = -c*D + mu*DD + r;
            u_0 = initial_condition(space, space_length);
            
            
        case 'wave_equation'

            c = parameter.c;
            rho = parameter.rho;
        
            right_hand_side_matrix = -(c^2 / rho) * [zeros(size(D)), zeros(size(D)); D, zeros(size(D))] - rho * [zeros(size(D)), D; zeros(size(D)), zeros(size(D))];
            u_0 = [initial_condition{1}(space, space_length), initial_condition{2}(space, space_length)];
            
        case 'burgers'
        
            mu = parameter.mu;
            right_hand_side_matrix = [];
            u_0 = initial_condition(space, space_length);
            
        otherwise
        
            error("%s is not a known example!", equation);
            
    end
    
    n_rows = n_vars * n_steps_space;
    POD_solution = zeros(n_rows, n_steps_time);
    
    reduced_state = (modes' * modes) \ (modes' * u_0');
    reduced_state_matrix = zeros(size(reduced_state, 1), n_steps_time);
    reduced_state_matrix(:, 1) = reduced_state;
    
    if (nargout > 3)
    
        if const_stepsize
            tsize = discretization.time.dt * (0:1:discretization.time.n_steps-1);
        else
            tsize = [0,1];
        end
    
        reduced_state_ode45 = reduced_state;
        reduced_state_ode23 = reduced_state;
        
        opts = odeset('Refine', 1);
        if (strcmp(equation, 'burgers'))
            tic;
            [t_ode45, reduced_state_ode45] = ode45(@(t, reduced_state) burgers(reduced_state, dx, modes, D, DD, mu), tsize, reduced_state_ode45, opts);
            computation_time_ode45 = toc;
            
            tic;
            [t_ode23, reduced_state_ode23] = ode23(@(t, reduced_state)burgers(reduced_state, dx, modes, D, DD, mu), tsize, reduced_state_ode23);
            computation_time_ode23 = toc;
        else
            tic;
            [t_ode45, reduced_state_ode45] = ode45(@(t, reduced_state) linear(reduced_state, modes, right_hand_side_matrix), tsize, reduced_state_ode45, opts);
            computation_time_ode45 = toc;
            
            tic;
            [t_ode23, reduced_state_ode23] = ode23(@(t, reduced_state)linear(reduced_state, modes, right_hand_side_matrix), tsize, reduced_state_ode23);
            computation_time_ode23 = toc;
        end
        
        POD_solution_ode45 = zeros(n_rows, size(reduced_state_ode45, 1));
        POD_solution_ode23 = zeros(n_rows, size(reduced_state_ode23, 1));
    end
    
    if (nargout > 3)
        for i = 1:size(reduced_state_ode45,1);
            POD_solution_ode45(:,i) = modes * reduced_state_ode45(i,:)';
        end
        for i = 1:size(reduced_state_ode23,1);
            POD_solution_ode23(:,i) = modes * reduced_state_ode23(i,:)';
        end
    end
    
    tic;
    for i = 1:n_steps_time-1;
        reduced_state = fsolve(@(next)implicit_trapezoid(equation, next, reduced_state, dt, dx, parameter, modes, right_hand_side_matrix, D, DD), reduced_state);
        reduced_state_matrix(:, i+1) = reduced_state;
    end
    computation_time = toc;
    
    for i = 1:n_steps_time
        POD_solution(:, i) = modes * reduced_state_matrix(:, i);
    end
    
end

function f = implicit_euler(equation, next, current, dt, dx, parameter, modes, right_hand_side_matrix, D, DD)
    if (strcmp(equation, 'burgers'))
        f = next - current - dt * (burgers(next, dx, modes, D, DD, parameter.mu));
    else
        f = next - current - dt * (linear(next, modes, right_hand_side_matrix));
    end
end

function f = implicit_trapezoid(equation, next, current, dt, dx, parameter, modes, right_hand_side_matrix, D, DD)
   
    if (strcmp(equation, 'burgers'))
        f = next - current - (dt / 2) * (burgers(next, dx, modes, D, DD, parameter.mu) + burgers(current, dx, modes, D, DD, parameter.mu));
    else
        f = next - current - (dt / 2) * (linear(next, modes, right_hand_side_matrix) + linear(current, modes, right_hand_side_matrix));
    end
    
end

function next = explicit_euler(equation, current, dt, dx, parameter, modes, right_hand_side_matrix, D, DD)
    if (strcmp(equation, 'burgers'))
        next = current + dt * (burgers(current, dx, modes, D, DD, parameter.mu));
    else
        next = current + dt * (linear(current, modes, right_hand_side_matrix));
    end
end

function f = burgers(current, dx, modes, D, DD, mu)
    f = (modes' * evaluate_burgers_right_hand_side(0, dx, modes * current, D, DD, mu));
end

function f = linear(current, modes, right_hand_side_matrix)
    f = modes' * right_hand_side_matrix * modes * current;
end
