function [shifted_POD_solution, reduced_state_matrix, computation_time, shifted_POD_solution_ode45, t_ode45, reduced_state_ode45, computation_time_ode45 shifted_POD_solution_ode23, t_ode23, reduced_state_ode23, computation_time_ode23] = shifted_pod_online(equation, modes, discretization, offline, parameter, every_mode_with_own_shift, use_equivariance, initial_condition, D, DD, const_stepsize);

    % Computes the shifted POD ROM solution based on a self written time integration (implicit trapezoidal rule) of the equation 'equation' and based on the parameters in the
    % structures discretization, offline, and parameter as well as the modes. Other inputs are finite difference matrices D and DD for the first and second  
    % spatial derivative, as well as a flag const_stepsize for the ode solvers.
    % 'every_mode_with_own_shift' and 'use_equivariance' are flags that tell the function that every mode should get its own shift / equivariance should be used.
    % If more than three outputs are requested, this function also uses matlab ode45 and ode23 solvers to compute the full order solution.

    %% data generation
    
    switch equation
        case 'advection_reaction_diffusion'
            n_vars = 1;
        case 'wave_equation'
            n_vars = 2;
        case 'burgers'
            n_vars = 1;
        otherwise
            error("%s is not a known example!", equation);
    end
    
    n_modes = offline.n_modes;
    n_total_modes = sum(n_modes);
    
    n_frames = offline.n_frames;
    shift_matrix_generator = offline.shift_matrix_generator;

    n_steps_space_per_var = size(modes, 1) / n_vars;
    
    length_space = discretization.space.length;
    dx = length_space / n_steps_space_per_var;
    
    space = dx * (1:n_steps_space_per_var);
    
    cfl = discretization.time.cfl;
    n_steps_time = discretization.time.n_steps;
    dt = discretization.time.dt;
    length_time = discretization.time.length;
    
    switch equation
        
        case 'advection_reaction_diffusion'
        
            c = parameter.c;
            mu = parameter.mu;
            r = parameter.r;
%              right_hand_side_matrix = -(c / (2*dx)) * D1 + (mu / dx^2) * D2 + r * speye(size(D1));
            right_hand_side_matrix = -c*D + mu*DD + r;
            u_0 = initial_condition(space, length_space);
            
        case 'wave_equation'
            
            c = parameter.c;
            rho = parameter.rho;
            right_hand_side_matrix = -(c^2 / rho) * [zeros(size(D)), zeros(size(D)); D, zeros(size(D))] - rho * [zeros(size(D)), D; zeros(size(D)), zeros(size(D))];
            u_0 = [initial_condition{1}(space, length_space), initial_condition{2}(space, length_space)];
            
        case 'burgers'
            
            mu = parameter.mu;
            right_hand_side_matrix = [];
            u_0 = initial_condition(space, length_space);
            
    end

    offset_vector = [0, cumsum(n_modes)];
    
    n_rows = n_vars * n_steps_space_per_var;
    shifted_POD_solution = zeros(n_rows, n_steps_time);
    
    if (every_mode_with_own_shift)
        reduced_state_matrix = zeros(n_total_modes + n_total_modes, n_steps_time);
        reduced_state = zeros(n_total_modes + n_total_modes, 1);
    else
        reduced_state_matrix = zeros(n_total_modes + n_frames, n_steps_time);
        reduced_state = zeros(n_total_modes + n_frames, 1);
    end
            
    reduced_state(1:n_total_modes) = (modes' * modes) \ (modes' * u_0');
    reduced_state_matrix(:,1) = reduced_state;
    
    % ODE 45 solver
    
    if (nargout > 3)
        reduced_state_ode45 = reduced_state;
        reduced_state_ode23 = reduced_state;
        
        opts = odeset('Refine', 1);
        
        if (strcmp(equation, 'burgers'))
        
            if const_stepsize
                tsize = discretization.time.dt * (0:1:discretization.time.n_steps-1);
            else
                tsize = [0,1];
            end
        
            tic;
            [t_ode45, reduced_state_ode45] = ode45(@(t, reduced_state) burgers(reduced_state, discretization, offline, parameter, modes, every_mode_with_own_shift, use_equivariance, n_vars, D, DD), tsize, reduced_state_ode45, opts);
            computation_time_ode45 = toc;
            
            tic;
            [t_ode23, reduced_state_ode23] = ode23(@(t, reduced_state)burgers(reduced_state, discretization, offline, parameter, modes, every_mode_with_own_shift, use_equivariance, n_vars, D, DD), tsize, reduced_state_ode23);
            computation_time_ode23 = toc;
        else
        
            if const_stepsize
                tsize = discretization.time.dt * (0:1:discretization.time.n_steps-1);
            else
                tsize = [0,1];
            end
        
            tic;
            [t_ode45, reduced_state_ode45] = ode45(@(t, reduced_state) linear(equation, reduced_state, discretization, offline, modes, right_hand_side_matrix, every_mode_with_own_shift, use_equivariance, D, DD), tsize, reduced_state_ode45, opts);
            computation_time_ode45 = toc;
            
            tic;
            [t_ode23, reduced_state_ode23] = ode23(@(t, reduced_state)linear(equation, reduced_state, discretization, offline, modes, right_hand_side_matrix, every_mode_with_own_shift, use_equivariance, D, DD), tsize, reduced_state_ode23);
            computation_time_ode23 = toc;
        end
        shifted_POD_solution_ode45 = zeros(n_rows, size(reduced_state_ode45, 1));
        shifted_POD_solution_ode23 = zeros(n_rows, size(reduced_state_ode23, 1));
    end
    
    %% Computation for reduced state for self-written time integration without adaptive stepsize.

    tic;
    for i = 1:n_steps_time-1
        
        
        reduced_state = fsolve(@(next)implicit_trapezoid(equation, next, reduced_state, dt, discretization, offline, parameter, modes, right_hand_side_matrix, every_mode_with_own_shift, use_equivariance, n_vars, D, DD), reduced_state);
        
        
        reduced_state_matrix(:,i+1) = reduced_state;
        
    end
    computation_time = toc;
    
    %% Build up of approximated solution

    for i = 1:n_steps_time
        
        if (every_mode_with_own_shift)
            
            for j = 1:n_total_modes
            
                current_shift = reduced_state_matrix(n_total_modes + j, i);
                current_shift_matrices = shift_matrix_generator(discretization, current_shift);

                if (n_frames > 1)
                    current_shift_matrix = kron(eye(n_frames), current_shift_matrices{1});
                else 
                    current_shift_matrix = current_shift_matrices{1};
                end

                shifted_POD_solution(:,i) = shifted_POD_solution(:,i) + current_shift_matrix * modes(:,j) * reduced_state_matrix(j,i);
                
            end
            
        else 
            
            for j = 1:n_frames
            
                current_shift = reduced_state_matrix(n_total_modes + j, i);
                current_shift_matrices = shift_matrix_generator(discretization, current_shift);
                if (n_frames > 1)
                    current_shift_matrix = kron(eye(n_frames), current_shift_matrices{1});
                else
                    current_shift_matrix = current_shift_matrices{1};
                end

                lower_index = offset_vector(j) + 1;
                upper_index = offset_vector(j + 1);
                
                shifted_POD_solution(:,i) = shifted_POD_solution(:,i) + current_shift_matrix * modes(:,lower_index:upper_index) * reduced_state_matrix(lower_index:upper_index, i);
                
            end
            
        end
        
    end
    
    % Build up of ode45 solution and ode23 solution.
    
    if (nargout > 3)
    
        % ode45
    
        for i = 1:size(reduced_state_ode45,1)
            
            if (every_mode_with_own_shift)
                
                for j = 1:n_total_modes
                
                    current_shift = reduced_state_ode45(i,n_total_modes + j);
                    current_shift_matrices = shift_matrix_generator(discretization, current_shift);
                    if (n_frames > 1)
                        current_shift_matrix = kron(eye(n_frames), current_shift_matrices{1});
                    else 
                        current_shift_matrix = current_shift_matrices{1};
                    end
                    
                    shifted_POD_solution_ode45(:,i) = shifted_POD_solution_ode45(:,i) + current_shift_matrix * modes(:,j) * reduced_state_ode45(i,j)';
                    
                end
                
            else 
                
                for j = 1:n_frames
                
                    current_shift = reduced_state_ode45(i,n_total_modes + j);
                    current_shift_matrices = shift_matrix_generator(discretization, current_shift);
                    if (n_frames > 1)
                        current_shift_matrix = kron(eye(n_frames), current_shift_matrices{1});
                    else
                        current_shift_matrix = current_shift_matrices{1};
                    end
                
                    lower_index = offset_vector(j) + 1;
                    upper_index = offset_vector(j + 1);
                    
                    shifted_POD_solution_ode45(:,i) = shifted_POD_solution_ode45(:,i) + current_shift_matrix * modes(:,lower_index:upper_index) * reduced_state_ode45(i, lower_index:upper_index)';
                    
                end
                
            end
            
        end
        
        % ode23
        
        for i = 1:size(reduced_state_ode23,1)
            
            if (every_mode_with_own_shift)
                
                for j = 1:n_total_modes
                
                    current_shift = reduced_state_ode23(i,n_total_modes + j);
                    current_shift_matrices = shift_matrix_generator(discretization, current_shift);
                    if (n_frames > 1)
                        current_shift_matrix = kron(eye(n_frames), current_shift_matrices{1});
                    else 
                        current_shift_matrix = current_shift_matrices{1};
                    end
                    
                    shifted_POD_solution_ode23(:,i) = shifted_POD_solution_ode23(:,i) + current_shift_matrix * modes(:,j) * reduced_state_ode23(i,j)';
                    
                end
                
            else 
                
                for j = 1:n_frames
                
                    current_shift = reduced_state_ode23(i,n_total_modes + j);
                    current_shift_matrices = shift_matrix_generator(discretization, current_shift);
                    if (n_frames > 1)
                        current_shift_matrix = kron(eye(n_frames), current_shift_matrices{1});
                    else
                        current_shift_matrix = current_shift_matrices{1};
                    end
                
                    lower_index = offset_vector(j) + 1;
                    upper_index = offset_vector(j + 1);
                    
                    shifted_POD_solution_ode23(:,i) = shifted_POD_solution_ode23(:,i) + current_shift_matrix * modes(:,lower_index:upper_index) * reduced_state_ode23(i, lower_index:upper_index)';
                    
                end
                
            end
            
        end
        
    end

end

function f = implicit_euler(equation, next, current, dt, discretization, offline, parameter, modes, right_hand_side_matrix, every_mode_with_own_shift, use_equivariance, n_vars, D, DD)
    if (strcmp(equation, 'burgers'))
        f = next - current - dt * (burgers(next, discretization, offline, parameter, modes, every_mode_with_own_shift, use_equivariance, n_vars, D, DD));
    else
        f = next - current - dt * (linear(equation, next, discretization, offline, modes, right_hand_side_matrix, every_mode_with_own_shift, use_equivariance, D, DD));
    end
end

function f = implicit_trapezoid(equation, next, current, dt, discretization, offline, parameter, modes, right_hand_side_matrix, every_mode_with_own_shift, use_equivariance, n_vars, D, DD)
   
    if (strcmp(equation, 'burgers'))
        f = next - current - (dt / 2) * (burgers(next, discretization, offline, parameter, modes, every_mode_with_own_shift, use_equivariance, n_vars, D, DD) + burgers(current, discretization, offline, parameter, modes, every_mode_with_own_shift, use_equivariance, n_vars, D, DD));
    else
        f = next - current - (dt / 2) * (linear(equation, next, discretization, offline, modes, right_hand_side_matrix, every_mode_with_own_shift, use_equivariance, D, DD) + linear(equation, current, discretization, offline, modes, right_hand_side_matrix, every_mode_with_own_shift, use_equivariance, D, DD));
    end
    
end

function next = explicit_euler(equation, current, dt, discretization, offline, parameter, modes, right_hand_side_matrix, every_mode_with_own_shift, use_equivariance, n_vars, D, DD)
    if (strcmp(equation, 'burgers'))
        next = current + dt * (burgers(current, discretization, offline, parameter, modes, every_mode_with_own_shift, use_equivariance, n_vars, D, DD));
    else
        next = current + dt * (linear(equation, current, discretization, offline, modes, right_hand_side_matrix, every_mode_with_own_shift, use_equivariance, D, DD));
    end
end

function f = burgers(current, discretization, offline, parameter, modes, every_mode_with_own_shift, use_equivariance, n_vars, D, DD)
    n_modes = offline.n_modes;
    n_total_modes = sum(n_modes);
    dx = discretization.space.dx;
    n_rows = n_vars * discretization.space.n_steps;
    mu = parameter.mu;
    n_frames = offline.n_frames;
    
    if (use_equivariance)
        W = compute_online_matrices_burgers_equivariance(current, discretization, offline, modes, every_mode_with_own_shift, D, DD);
    else
        W = compute_online_matrices_burgers(current, discretization, offline, modes, every_mode_with_own_shift, D, DD);
    end
    
    if (every_mode_with_own_shift)
        ansatz = [W(:, 1:n_total_modes), zeros(n_rows, n_total_modes)] * current;
    else
        ansatz = [W(:, 1:n_total_modes), zeros(n_rows, n_frames)] * current;
    end
    
    left_online_matrix = W'*W;
    f = left_online_matrix \ (W'*evaluate_burgers_right_hand_side(0, dx, ansatz, D, DD, mu));
end

function f = linear(equation, current, discretization, offline, modes, right_hand_side_matrix, every_mode_with_own_shift, use_equivariance, D, DD)
    % For the wave equation, we cannot use equivariance.
    if (use_equivariance && ~strcmp(equation, 'wave-equation'))
        [left_online_matrix, right_online_matrix] = compute_online_matrices_equivariance(current, discretization, offline, modes, right_hand_side_matrix, every_mode_with_own_shift, D, DD);
    else
        [left_online_matrix, right_online_matrix] = compute_online_matrices(current, discretization, offline, modes, right_hand_side_matrix, every_mode_with_own_shift, D, DD);
    end
    f = left_online_matrix \ (right_online_matrix * current);
end
