%% Projection-Based Model Reduction with Dynamically Transformed Modes
% This is the main file to compute the numerical examples for the paper.
%
% Reference:
%   F.Black, P. Schulze, B. Unger: Projection-Based Model Reduction with Dynamically Transformed Modes (2020)
%
% Authors:
%   Felix Black, black@math.tu-berlin.de
%   Philipp Schulze, pschulze@math.tu-berlin.de
%   Benjamin Unger, unger@math.tu-berlin.de
%
% Version:
%   1.0 (June 21, 2020)
% 
clear
close all
clc

%--------------------------------------------------------------------------
%% Setup
%--------------------------------------------------------------------------

% Possible examples

% - advection_reaction_diffusion: create a full chain for the advection reaction diffusion equation, i.e., compute the full order solution, determine modes based on that snapshot, and solve the reduced order model
% - wave_equation: create a full chain for the wave equation, i.e., compute the full order solution, determine modes based on that snapshot, and solve the reduced order model
% - burgers: create a full chain for the wave equation, i.e., compute the full order solution, determine modes based on that snapshot, and solve the reduced order model

% Examples for the reproduction of the results in the paper

% - parameter_sampling_advection_diffusion: WARNING: This example is computationally expensive and may take some time. Determines POD and shifted POD modes from a full order solution with velocity parameter c = 1. Then creates full order models, POD reduced order models and shifted POD reduced order models for different velocity parameters. Compares the relative online error in L2 norm for all generated ROMs and the computation times. 

% single_transformation_vs_multiple_transformations: This example compares the shifted POD relative online error that is produced if a single transformation is chosen for all modes vs the error if every mode gets its own transformation. The error is compared columnwise.

% wave_equation_ode_solver: A comparison of the time steps needed by the ode45 and ode23 solver to solve the full order model, a POD reduced order model with 14 modes and a shifted POD reduced order model with 2 modes for the wave_equation, with velocity parameter c = 1 and density parameter rho = 1.

% viscous_burgers_equation: A comparison of the relative L2 offline and online error for the POD reduced order model with 7 and 32 modes and the shifted POD reduced order model with 7 modes.

example = 'advection_reaction_diffusion';
%  example = 'wave_equation';
%  example = 'burgers';
%  example = 'parameter_sampling_advection_diffusion';
%  example = 'single_transformation_vs_multiple_transformations';
%  example = 'wave_equation_ode_solver';
%  example = 'viscous_burgers_equation';

%--------------------------------------------------------------------------
%% User input
%--------------------------------------------------------------------------

addpath('HELPER');
addpath('INITIAL_CONDITION');
addpath('SHIFT');
addpath('COST_FUNCTIONAL');
addpath('HELPER');
addpath('DATA_GENERATOR');
addpath('SPOD');
addpath('SPOD_ONLINE');
addpath('INTERPOLATION');
addpath('PLOT');
addpath('POD');
addpath('ERROR');

switch (example)

    case 'advection_reaction_diffusion'

        % Velocity parameter, can be specified as a single scalar, or a row vector of scalar values
        % Examples: c_values = 1, c_values = [0.5, 1, 1.5], c_values = (-5:0.5:5);
        % For every specified value, one full chain from FOM to ROM is run.
        c_values = 1;
        
        % Velocity parameter, can be specified as a single scalar, or a row vector of scalar values
        % Examples: mu_values = 1, mu_values = [0.5, 1, 1.5], mu_values = (-5:0.5:5);
        % For every specified value, one full chain from FOM to ROM is run.
        mu_values = 0.002;
        
        % Reaction parameter, can be specified as a single scalar, or a row vector of scalar values
        % Examples: r_values = 1, r_values = [0.5, 1, 1.5], r_values = (-5:0.5:5);
        % For every specified value, one full chain from FOM to ROM is run.
        r_values = 0;
        
        % Values for number of steps in space and time, specify as a whole positive number.
        % Examples: discretization_default.space.n_steps = 300, discretization_default.time.n_steps = 50
        discretization_default.space.n_steps = 200;
        discretization_default.time.n_steps = 200;
        
        % The number of modes for shifted POD, specify as a whole positive number.
        % Examples: For pure advection, i.e., mu = 0, r = 0: offline_default.n_modes = 1;
        % For advection-diffusion-reaction: offline_default.n_modes = 2;
        
        offline_default.n_modes = 2;
        
        % Do not touch
        
        % FOM parameters
        fom = create_fom_struct(example, 'multiplicative', 'parameter.c', c_values, 'parameter.mu', mu_values, 'parameter.r', r_values);

        % Discretization parameters
        discretization = create_discretization_struct(example, fom, discretization_default);
        
        % Offline phase parameters
        offline = create_offline_struct(example, fom, discretization, offline_default);

        % ROM parameters
        rom = create_rom_struct(example, 'parameter.c', c_values, 'parameter.mu', mu_values, 'parameter.r', r_values);
        
        clear discretization_default;
        clear offline_default;
        
    case 'wave_equation'
        
        % Velocity parameter, can be specified as a single scalar, or a row vector of scalar values
        % Examples: c_values = 1, c_values = [0.5, 1, 1.5], c_values = (-5:0.5:5);
        % For every specified value, one full chain from FOM to ROM is run.
        c_values = 1;
        
        % Density parameter, can be specified as a single scalar, or a row vector of scalar values
        % Examples: rho_values = 1, rho_values = [0.5, 1, 1.5], rho_values = (-5:0.5:5);
        % For every specified value, one full chain from FOM to ROM is run.
        rho_values = 1;
        
        % Values for number of steps in space and time, specify as a whole positive number.
        % Examples: discretization_default.space.n_steps = 300, discretization_default.time.n_steps = 50
        discretization_default.space.n_steps = 200;
        discretization_default.time.n_steps = 200;
        
        % The number of modes for shifted POD, specify as a 1 by 2 array of whole positive numbers.
        % Examples: offline_default.n_modes = [1, 1] is likely to give the best results
        % offline_default.n_modes = [2, 2];
        offline_default.n_modes = [1, 1];
        
        % Do not touch
        
        % FOM parameters
        fom = create_fom_struct(example, 'multiplicative', 'parameter.c', c_values, 'parameter.rho', rho_values);
        
        % Discretization parameters
        discretization = create_discretization_struct(example, fom, discretization_default);
        
        % Offline phase parameters
        offline = create_offline_struct(example, fom, discretization, offline_default);

        % ROM parameters
        rom = create_rom_struct(example, 'parameter.c', c_values, 'parameter.rho', rho_values);
        
        clear discretization_default;
        clear offline_default;
        
    case 'burgers'

        % Velocity parameter, can be specified as a single scalar, or a row vector of scalar values
        % Examples: mu_values = 1, mu_values = [0.5, 1, 1.5], mu_values = (-5:0.5:5);
        % For every specified value, one full chain from FOM to ROM is run.
        mu_values = [0.002, 0.004, 0.006];
        
        % Values for number of steps in space and time, specify as a whole positive number.
        % Examples: discretization_default.space.n_steps = 300, discretization_default.time.n_steps = 50
        discretization_default.space.n_steps = 200;
        discretization_default.time.n_steps = 200;
        
        % The number of modes for shifted POD, specify as a whole positive number.
        % Examples: offline_default.n_modes = 5, offline_default.n_modes = 10
        offline_default.n_modes = 7;
        
        % FOM parameters
        fom = create_fom_struct(example, 'multiplicative', 'parameter.mu', mu_values);

        % Discretization parameters
        discretization = create_discretization_struct(example, fom, discretization_default);
        
        % Offline phase parameters
        offline = create_offline_struct(example, fom, discretization, offline_default);

        % ROM parameters
        rom = create_rom_struct(example, 'parameter.mu', mu_values);
        
        clear discretization_default;
        clear offline_default;

    case 'parameter_sampling_advection_diffusion'
    
        % Velocity parameter, can be specified as a single scalar, or a row vector of scalar values
        % Examples: c_values = 1, c_values = [0.5, 1, 1.5], c_values = (-5:0.5:5);
        % For every specified value, one full chain from FOM to ROM is run, but the ROMs
        % are based on the modes that are detected from the samples.
        c_values = (-5:0.1:5);
        
        % Choose c = 1 as sample
        sample_values = c_values(c_values == 1);
        samples = find(c_values == sample_values);
        
        % Values for number of steps in space and time, specify as a whole positive number.
        % Examples: discretization_default.space.n_steps = 300, discretization_default.time.n_steps = 50
        discretization_default.space.n_steps = 200;
        discretization_default.time.n_steps = 200;
        
        % The number of modes for shifted POD, specify as a whole positive number.
        % WARNING: since this example contains diffusion, choose n_modes >= 2.
        offline_default.n_modes = 2;

        % Do not touch
        
        % FOM parameters
        fom = create_fom_struct('advection_reaction_diffusion', 'multiplicative', 'parameter.c', c_values);
        
        % Discretization parameters
        discretization = create_discretization_struct('advection_reaction_diffusion', fom, discretization_default);
        
        % Offline phase parameters
        offline = create_offline_struct('advection_reaction_diffusion', fom, discretization, offline_default);
        
        % ROM parameters
        rom = create_rom_struct('advection_reaction_diffusion', 'parameter.c', c_values);
        
        clear discretization_default;
        clear offline_default;
    
    case 'single_transformation_vs_multiple_transformations'
    
        % Only for reconstruction of the example in the paper, do not touch.
        
        discretization_default.space.n_steps = 200;
        discretization_default.time.n_steps = 200;
        
        offline_default.n_modes = 2;
        
        fom = create_fom_struct('advection_reaction_diffusion', 'multiplicative');
        discretization = create_discretization_struct('advection_reaction_diffusion', fom, discretization_default);
        offline = create_offline_struct('advection_reaction_diffusion', fom, discretization, offline_default);
        rom = create_rom_struct('advection_reaction_diffusion');
        
        clear discretization_default;
        clear offline_default;
        
    case 'wave_equation_ode_solver'
    
        % Only for reconstruction of the example in the paper, do not touch.
    
        discretization_default.space.n_steps = 200;
        discretization_default.time.n_steps = 200;
        
        offline_default.n_modes = [1, 1];
        
        fom = create_fom_struct('wave_equation', 'multiplicative');
        discretization = create_discretization_struct('wave_equation', fom, discretization_default);
        offline = create_offline_struct('wave_equation', fom, discretization, offline_default);
        rom = create_rom_struct('wave_equation');
        
        clear discretization_default;
        clear offline_default;
        
    case 'viscous_burgers_equation'
    
        % Only for reconstruction of the example in the paper, do not touch.
    
        discretization_default.space.n_steps = 200;
        discretization_default.time.n_steps = 200;
        
        offline_default.n_modes = 7;
        
        fom = create_fom_struct('burgers', 'multiplicative');
        discretization = create_discretization_struct('burgers', fom, discretization_default);
        offline = create_offline_struct('burgers', fom, discretization, offline_default);
        rom = create_rom_struct('burgers');
        
        clear discretization_default;
        clear offline_default;
        
    otherwise
        error("Example '%s' not specified.", example);
end

%-------------------------------------------------------------------------------
%% Calculations
%-------------------------------------------------------------------------------

fprintf('Example: %s\n\n',example);

switch example

    case 'advection_reaction_diffusion'
    
        % Compute FOM solution with ode45 and force ode45 to return at every grid point
        [~, ~, snapshot_matrix_ode45, ~, ~, ~, ~, ~] = compute_truth_solution_struct_array(fom, discretization, 1);
        
        % Determine modes
        [modes, ~, relative_L2_offline_error] = shifted_pod_offline_struct_array(snapshot_matrix_ode45, fom, discretization, offline, 'normal');
        
        % Compute shifted pod solution with ode45 and force ode45 to return at every grid point
        [~, ~, ~, shifted_pod_solution_ode45, ~, ~, ~, ~, ~, ~, ~] = shifted_pod_online_struct_array(rom, discretization, offline, modes, 0, 0, 1);

        % Compute relative L2 online error
        relative_L2_online_error = compute_relative_L2_online_error(snapshot_matrix_ode45, shifted_pod_solution_ode45);
        
        
    case 'wave_equation'
    
        % Compute FOM solution with ode45 and force ode45 to return at every grid point
        [~, ~, snapshot_matrix_ode45, ~, ~, ~, ~, ~] = compute_truth_solution_struct_array(fom, discretization, 1);
        
        % Determine modes
        [modes, ~, relative_L2_offline_error] = shifted_pod_offline_struct_array(snapshot_matrix_ode45, fom, discretization, offline, 'normal');
        
        % Compute shifted pod solution with ode45 and force ode45 to return at every grid point
        [~, ~, ~, shifted_pod_solution_ode45, ~, ~, ~, ~, ~, ~, ~] = shifted_pod_online_struct_array(rom, discretization, offline, modes, 0, 0, 1);

        % Compute relative L2 online error
        relative_L2_online_error = compute_relative_L2_online_error(snapshot_matrix_ode45, shifted_pod_solution_ode45);
        
    case 'burgers'
    
        % Compute FOM solution with ode45 and force ode45 to return at every grid point
        [~, ~, snapshot_matrix_ode45, ~, ~, ~, ~, ~] = compute_truth_solution_struct_array(fom, discretization, 1);
        
        % Determine modes
        [modes, ~, relative_L2_offline_error] = shifted_pod_offline_struct_array(snapshot_matrix_ode45, fom, discretization, offline, 'normal');
        
        % Compute shifted pod solution with ode45 and force ode45 to return at every grid point
        [~, ~, ~, shifted_pod_solution_ode45, ~, ~, ~, ~, ~, ~, ~] = shifted_pod_online_struct_array(rom, discretization, offline, modes, 0, 0, 1);

        % Compute relative L2 online error
        relative_L2_online_error = compute_relative_L2_online_error(snapshot_matrix_ode45, shifted_pod_solution_ode45);

    case 'parameter_sampling_advection_diffusion'
        
        % Make n_runs to average the computation times
        n_runs = 1;
        computation_time_fom_matrix = zeros(n_runs, length(c_values));
        computation_time_rom_matrix = zeros(n_runs, length(c_values));
        computation_time_pod_3_matrix = zeros(n_runs, length(c_values));
        computation_time_pod_11_matrix = zeros(n_runs, length(c_values));
        
        for i = 1:n_runs
            % Compute FOM with ode45 and force ode45 to return at every grid point
            [~, ~, snapshot_matrix_ode45, ~, computation_time_fom_matrix(i,:), ~, ~, ~] = compute_truth_solution_struct_array(fom, discretization, 1);
        end
            
        % Run shifted POD offline phase in sampling mode
        [modes, ~, relative_L2_offline_error] = shifted_pod_offline_struct_array(snapshot_matrix_ode45(samples), fom(samples), discretization(samples), offline, 'sampling');
        
        % POD offline phase with 3 and 11 modes
        [pod_modes_3, pod_3_relative_L2_offline_error] = POD_offline_struct_array(snapshot_matrix_ode45(samples), 3, 'sampling');
        [pod_modes_11, pod_11_relative_L2_offline_error] = POD_offline_struct_array(snapshot_matrix_ode45(samples), 11, 'sampling');
        
        for i = 1:n_runs
        
            % Run online phase depending on the modes yielded by the sampling, but this time use equivariance
            [~, ~, ~, shifted_pod_solution_ode45_equi, ~, ~, computation_time_rom_matrix(i,:), ~, ~, ~, ~] = shifted_pod_online_struct_array(rom, discretization, offline, modes, 0, 1, 1);

            % Compute POD solution

            [~, ~, ~, pod_solution_ode45_3, ~, ~, computation_time_pod_3_matrix(i,:), ~, ~, ~, ~] = POD_online_struct_array(rom, discretization, pod_modes_3, 1);
            [~, ~, ~, pod_solution_ode45_11, ~, ~, computation_time_pod_11_matrix(i,:), ~, ~, ~, ~] = POD_online_struct_array(rom, discretization, pod_modes_11, 1);
        
        end
        
        % Computation of online errors
        
        pod_3_relative_L2_online_error = compute_relative_L2_online_error(snapshot_matrix_ode45, pod_solution_ode45_3);
        pod_11_relative_L2_online_error = compute_relative_L2_online_error(snapshot_matrix_ode45, pod_solution_ode45_11);
        
        relative_L2_online_error_equi = compute_relative_L2_online_error(snapshot_matrix_ode45, shifted_pod_solution_ode45_equi);
        
        % Compute medians of the computation times
        median_computation_time_fom = median(computation_time_fom_matrix);
        median_computation_time_rom = median(computation_time_rom_matrix);
        median_computation_time_pod_3 = median(computation_time_pod_3_matrix);
        median_computation_time_pod_11 = median(computation_time_pod_11_matrix);
        
        % Plot results
        plot_parameter_sampling_advection_diffusion;
        
    case 'single_transformation_vs_multiple_transformations'
    
        % Compute FOM with self written time integration
        [snapshot_matrix, ~] = compute_truth_solution_struct_array(fom, discretization, 1);
        
        % Determine modes
        [modes, ~, relative_L2_offline_error] = shifted_pod_offline_struct_array(snapshot_matrix, fom, discretization, offline, 'normal');
        
        % Compute shifted pod solution, and use the same transformation for every mode
        [shifted_pod_solution_one_trafo, reduced_state_matrix_one_trafo, ~] = shifted_pod_online_struct_array(rom, discretization, offline, modes, 0, 0, 1);
        
        % Compute shifted pod solution and use one transformation per mode
        [shifted_pod_solution_multiple_trafo, reduced_state_matrix_multiple_trafo, ~] = shifted_pod_online_struct_array(rom, discretization, offline, modes, 1, 0, 1);
        
        % Compute errors
        relative_L2_online_error_one_trafo = relative_L2_norm(snapshot_matrix{1}, snapshot_matrix{1} - shifted_pod_solution_one_trafo{1});
        relative_L2_online_error_multiple_trafo = relative_L2_norm(snapshot_matrix{1}, snapshot_matrix{1} - shifted_pod_solution_multiple_trafo{1});
        
        % Plot results
        plot_single_transformation_vs_multiple_transformations;
        
    case 'wave_equation_ode_solver'
    
        % Compute FOM with ode45 and ode23
        [snapshot_matrix, ~, snapshot_matrix_ode45, t_ode45, ~, snapshot_matrix_ode23, t_ode23, ~] = compute_truth_solution_struct_array(fom, discretization, 0);
        
        % Determine modes
        [modes, ~, relative_L2_offline_error] = shifted_pod_offline_struct_array(snapshot_matrix, fom, discretization, offline, 'normal');
        
        % Compute shifted pod solution with ode45 and ode23
        [shifted_pod_solution, ~, ~, shifted_pod_solution_ode45, t_ode45_rom, ~, ~, shifted_pod_solution_ode23, t_ode23_rom, ~, ~] = shifted_pod_online_struct_array(rom, discretization, offline, modes, 0, 0, 0);
        
        % Determine modes for POD
        [pod_modes, pod_relative_L2_offline_error] = POD_offline_struct_array(snapshot_matrix, 14, 'normal');
        
        % Compute POD solution with ode45 and ode23
        [pod_solution, ~, ~, pod_solution_ode45, t_ode45_pod, ~, ~, pod_solution_ode23, t_ode23_pod, ~, ~] = POD_online_struct_array(rom, discretization, pod_modes, 0);
        
        % Compute errors
        relative_L2_online_error = relative_L2_norm(snapshot_matrix{1}, snapshot_matrix{1} - shifted_pod_solution{1});
        pod_relative_L2_online_error = relative_L2_norm(snapshot_matrix{1}, snapshot_matrix{1} - pod_solution{1});
        
        % Plot results
        plot_wave_equation_ode_solver;
        
    case 'viscous_burgers_equation'
    
        % Compute FOM solution
        [snapshot_matrix, ~, ~] = compute_truth_solution_struct_array(fom, discretization, 1);
        
        % Determine modes
        [modes, ~, relative_L2_offline_error] = shifted_pod_offline_struct_array(snapshot_matrix, fom, discretization, offline, 'normal');
        
        % Compute shifted pod solution
        [shifted_pod_solution, reduced_state_matrix, ~] = shifted_pod_online_struct_array(rom, discretization, offline, modes, 0, 0, 1);
        
        % Determine POD modes
        [pod_modes_7, pod_relative_L2_offline_error_7] = POD_offline_struct_array(snapshot_matrix, 7, 'normal');
        [pod_modes_32, pod_relative_L2_offline_error_32] = POD_offline_struct_array(snapshot_matrix, 32, 'normal');
        
        % Compute POD solutions
        [pod_solution_7, ~, ~] = POD_online_struct_array(rom, discretization, pod_modes_7, 0);
        [pod_solution_32, ~, ~] = POD_online_struct_array(rom, discretization, pod_modes_32, 0);
        
        % Compute errors
        relative_L2_online_error = relative_L2_norm(snapshot_matrix{1}, snapshot_matrix{1} - shifted_pod_solution{1});
        pod_relative_L2_online_error_7 = relative_L2_norm(snapshot_matrix{1}, snapshot_matrix{1} - pod_solution_7{1});
        pod_relative_L2_online_error_32 = relative_L2_norm(snapshot_matrix{1}, snapshot_matrix{1} - pod_solution_32{1});
        
        % Plot results
        plot_viscous_burgers_equation;
        
    otherwise
    
        error("Example '%s' not specified.", example);
        
end
