%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% This function computes the shift-matrices for a fixed spatial dimension
% and a fixed transport velocity for periodic boundary conditions. Inputs
% are the "discretization"-structure, which contains all relevant parameters, and
% the variable shift, which is, depending on the problem, a tensor, a matrix,
% or a vector (at the moment, shift will be a vector, since transport-velocity
% and spatial dimension are fixed).
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [shift_matrices, shift_matrices_derivative]  = shift_matrix_generator_periodic(discretization, shift)

% assumed periodicity: 'zero'th value = last value (boundary is only
% represented once in the snapshot matrix).

% assumes that interpolationStencilIndices = [-1 0 1 2]:

n_steps_space = discretization.space.n_steps;  

dx = discretization.space.dx; 

n_shifts = length(shift);

shift_matrices = cell(n_shifts, 1); 

% Initialize the diagonal index-vector as a 1 times 4 empty row vector. 

diagonal_index_vector = NaN(1, 4);

% If the call to this function also specifies that the derivatives of the
% shift-matrices are needed ... 

if (nargout > 1)
    
    shift_matrices_derivative = cell(n_shifts, 1); 
    
    for i = 1:n_shifts
        % Set the current shift length.
        shift_length = shift(i);
        % If the shift length is not a multiple of the spatial step size, we need to determine the correct stencil. Basically, we need to interpolate. With this goal in mind, since we are using a forward stencil, we set the stencil index at position 2 (corresponding to the current step) to a multiple of the spatial step size.
        diagonal_index_vector(2) = floor(shift_length/dx);
    
        % The entry for the last step is just the entry for the current step minus one.
        diagonal_index_vector(1) = diagonal_index_vector(2)-1;
    
        % The entry for the next step is just the entry of the current step plus one.
        diagonal_index_vector(3) = diagonal_index_vector(2)+1;
    
        % The entry for the step after the next step is just the current one plus 2.
        diagonal_index_vector(4) = diagonal_index_vector(2)+2;
        
        dx_weight = shift_length / dx - diagonal_index_vector(2);
        
        % First, convert the diagonal index vector to a vector of absolute values. Create a logical array (meaning an array of boolean values), where each entry indicates if the number in diagonal index vector is greater than nStepsSpace - 1. If this condition is true for an entry in diagonal index vector, the boolean array will have a 1 at that position, indicating that the corresponding entry in the index vector exceeds the number of diagonals that are feasible.
        out_of_matrix_diagonal_indices = abs(diagonal_index_vector) > n_steps_space - 1;
        
        % Create a vector, containing only the entries of diagonal index vector that are out of bounds.
        out_of_matrix_diagonal_values = diagonal_index_vector(out_of_matrix_diagonal_indices);
        
        % Modify the entries that are out of bounds: Take the original entries and take them modulo the number of spatial steps times their sign. This operation gives the correct diagonals for periodic boundary conditions.
        diagonal_index_vector(out_of_matrix_diagonal_indices) = mod(out_of_matrix_diagonal_values, n_steps_space * sign(out_of_matrix_diagonal_values));

        %After the operations above, diagonal index vector should only contain values from -nStepsSpace+1 to nStepsSpace-1. To get the dual index vector, nStepsSpace is either added or subtracted, depending on the sign of the entry in the index vector. A Problem occurs when an entry in diagonal index vector is zero: It stays zero if one does the above described operation, since the sign of zero is zero. But then, when building up the variable bands below, bands contains zero as a diagonal twice, which we do not want. That is why we add 0.1 to diagonal index vector in the sign, to achieve a positive sign for a zero entry in diagonal index vector. The result for a zero value is then nStepsSpace, which is ignored in the call to spdiags below, since the diagonal does not exist.
        dual_diagonal_index_vector = diagonal_index_vector-(n_steps_space) * sign(diagonal_index_vector + 0.1);
        
        % Form the bands where the shiftMatrices need values.
        bands = [diagonal_index_vector dual_diagonal_index_vector];
        
        % Compute a lagrange interpolation for the non-integer rest.
        Lagrange = lagrange(dx_weight);
        
        % Compute the entries for the bands from the lagrange interpolation.
        entries = repmat(Lagrange', n_steps_space, 2);
        
        % Build the shift matrix.
        shift_matrices{i} = spdiags(entries, bands, n_steps_space, n_steps_space);
        
        % Compute the derivative of the lagrange interpolation above.
        Lagrange_diff = lagrange_derivative(dx_weight) / dx;
        % Compute entries for the shift matrix derivative.
        entries = repmat(Lagrange_diff', n_steps_space, 2);
        % Build the shift matrix derivative.
        shift_matrices_derivative{i} = spdiags(entries, bands, n_steps_space, n_steps_space);   
    end
    
% Otherwise, if only the shift-matrices are needed, then ...    
    
else
    
    % ... for every time-step, compute ...
    
    for i=1:n_shifts
        
        % ... the shift-length as the entry in cs at position i
        
        shift_length = shift(i);  
        
        % ... the entry at position 2 of the diagonal index-vector as the
        % shift-length divided by the length of one spatial step, rounded
        % down
        
        diagonal_index_vector(2) = floor(shift_length / dx);
        
        % ... the other entries of the diagonal index-vector by the
        % assumed interpolation stencil.
        
        diagonal_index_vector(1) = diagonal_index_vector(2) - 1;
        diagonal_index_vector(3) = diagonal_index_vector(2) + 1;  
        diagonal_index_vector(4) = diagonal_index_vector(2) + 2;
        
        % ... the weight of dxi as shift-length divided by the
        % length of one spatial step minus the entry at position 2 in the
        % diagonal index-vector (i.e. dxiWeight is the rest that is
        % neglected by rounding down when computing diagonal_index_vector(2) 
        % above). 
        
        dx_weight = shift_length / dx - diagonal_index_vector(2);
        
        % ... check if the indices in the diagonal index-vector are within
        % the bounds of the snapshot matrix. outOfMatrixDiagonalIndices is a vector
        % containing either a 1 if the index at this position is out of
        % bounds, or a 0 if it is not.
        
        out_of_matrix_diagonal_indices = abs(diagonal_index_vector) > n_steps_space - 1;
        
        % Cut out the indices that are out of bounds and put those in the
        % vector "outOfMatrixdiagonal_index_vector".
        
        out_of_matrix_diagonal_values = diagonal_index_vector(out_of_matrix_diagonal_indices);
        
        % Since we assume periodic boundary conditions, compute the actual
        % index for the indices that are out of bounds by taking the modulo
        
        diagonal_index_vector(out_of_matrix_diagonal_indices) = mod(out_of_matrix_diagonal_values, n_steps_space * sign(out_of_matrix_diagonal_values)) ;
        
        % Compute the dual diagonal index-vector.
        
        dual_diagonal_index_vector = diagonal_index_vector - (n_steps_space) * sign(diagonal_index_vector + 0.1);
        
        % Set the "bands"-vector as the concatenation of diagonal_index_vector
        % and dualdiagonal_index_vector.
        
        bands = [diagonal_index_vector dual_diagonal_index_vector];
        
        %Compute values of Lagrange-polynomials at dxiWeight.
        
        Lagrange = lagrange(dx_weight);
        
        % Compute the entries for the shift-matrix. It has 2 columns. In
        % each column, it has the values of the transposed lagrange-vector
        % stacked Nxi times on top of each other.
        
        entries = repmat(Lagrange', n_steps_space, 2);
        
        % Set the shift-matrix as a sparse diagonal matrix of size Nxi
        % times Nxi, with the nonzero diagonals that are specified by
        % bands. The nonzero diagonals are the columns of the
        % entries-matrix.
        
        shift_matrices{i} = spdiags(entries, bands, n_steps_space, n_steps_space);   
    end
end
