function [modes, offline, min_rel_error] = shifted_pod_offline_struct_array(snapshot_matrix, fom, discretization, offline, behaviour)

    % Runs the shifted POD offline phase for all snapshot matrices contained in the variable snapshot_matrix, which is assumed 
    % to be a cell, based on the parameters contained in the structure arrays fom and discretization and the offline struct.
    % The function has two modes which can be set via the behaviour variable:
    % either behaviour = 'normal', or 'sampling'.
    % 'normal' means that for each snapshot_matrix, n_modes modes are computed, 'sampling' means that
    % all snapshot matrices are concatenated and n_modes averaged modes are computed.

switch behaviour
    
    case 'normal'
    
        n_samples = length(snapshot_matrix);
        modes = cell(1, n_samples);
        min_rel_error = zeros(1, n_samples);
        
        for i = 1:n_samples
            [modes{i}, ~, min_rel_error(i)] = shifted_pod_offline(fom(i).equation, snapshot_matrix{i}, discretization(i), offline, fom(i).parameter);
        end
    
    case 'sampling'
    
        % Concatenate all snapshot matrices into one big matrix
        n_samples = length(snapshot_matrix);
        n_steps_time_per_sample = zeros(1, n_samples);
        dt_per_sample = zeros(1, n_samples);
        
        for i = 1:n_samples
            n_steps_time_per_sample(i) = size(snapshot_matrix{i}, 2);
            dt_per_sample(i) = discretization(i).time.dt;
        end
        
        snapshot_matrix = cell2mat(snapshot_matrix);
        n_steps_space = size(snapshot_matrix, 1);
        n_steps_space_per_var = discretization(1).space.n_steps;
        n_vars = n_steps_space / n_steps_space_per_var;
        
        time_per_sample = cell(1, n_samples);
        for i = 1:n_samples
            final = n_steps_time_per_sample(i)-1;
            time_per_sample{i} = dt_per_sample(i) * (0:1:final);
        end
        
        time = cell2mat(time_per_sample);
        
        switch fom(1).equation
            
            case 'advection_reaction_diffusion'
%                  offline.n_frames = 1;
                
                c_per_sampling = zeros(1, n_samples);
                for i = 1:n_samples
                    c_per_sampling(i) = fom(i).parameter.c;
                end
                
                shift = cell(1,n_samples);
                for i = 1:n_samples
                    shift{i} = time_per_sample{i}' * c_per_sampling(i);
                end
                shift = cell2mat(shift);
                shift = reshape(shift, [], 1);
                
%                  offline.shift_length_detection.velocity = c_per_sampling;
%                  shift = time' .* c_per_sampling;
%                  shift = reshape(shift, [], 1);
                
            case 'wave_equation'
%                  offline.n_frames = 2;
                
                c_per_sampling = zeros(1, n_samples);
                for i = 1:n_samples
                    c_per_sampling(i) = fom(i).parameter.c;
                end
                
                c_pos_neg = zeros(1, 2*n_samples);
                c_pos_neg(1:2:end) = c_per_sampling;
                c_pos_neg(2:2:end) = -c_per_sampling;
                
%                  offline.shift_length_detection.velocity = c_pos_neg;
                shift = time' .* c_pos_neg;
                
            case 'burgers'
%                  offline.n_frames = 1;
                
                % Check the dependence of cost functional on discretization
                [shift, shift_derivative] = steepest_descent(offline.path_detection.starting_value, @(shift)offline.path_detection.cost_functional(shift, discretization(1), offline, snapshot_matrix), offline.path_detection.max_iterations, offline.path_detection.tolerance);
                
                if(isscalar(shift))
                    shift = shift * time';
                end
                
                % Correct a possible offset such that p(0) = 0. Otherwise, the modes and the path are not compatible, since the modes are "in the wrong position" (i.e. the offset is not incorporated in the modes). This leads to bugs in the online phase.
                shift = shift - shift(1);
                
        end

        subspace_decomposition = @subspace_decomposition_by_shifted_POD_iteration;

        [shifted_POD_system, shifted_snapshot_matrix, rel_errors] = subspace_decomposition(discretization(1), offline, snapshot_matrix, shift);
        
        min_rel_error = min(rel_errors);

        modes = zeros(n_steps_space, sum(offline.n_modes));

        helper = [0, cumsum(offline.n_modes)];

        for i = 1:length(offline.n_modes)
            modes(:,helper(i)+1:helper(i+1)) = shifted_POD_system.U{i};
        end
        
        % To comply with the rest of the code, make modes a cell.
        temp = modes;
        modes = cell(1);
        modes{1} = temp;
        
    otherwise
        error("Behaviour %s is not a known behaviour for this method.", behaviour);
end

end
